import furl
import logging
from random import randint
import requests
from requests.exceptions import RequestException
import time

logger = logging.getLogger(__name__)

SUCCESS_HTTP_CODE = [200, 201, 204]


class HTTPErrorCode(Exception):
    """Exception raised by a failed call on Failed HTTP Requests"""


def retry_http(call):
    """Wrapper used to retry HTTP Errors with an exponential backoff

    Args:
        call: function being wrapped

    Returns:
        the wrapped function
    """

    def _retry_http(*args, **kwargs):
        """Retry a function call when catching requests.exceptions.RequestException

        """
        last_exception = None
        multiplier = 1.5
        retry_interval = 0.5
        randomization_factor = 0.5
        total_sleep_time = 0
        # Capped to 10 seconds
        max_sleep_time = 10

        request_nb = 0
        while total_sleep_time < max_sleep_time:
            try:
                return call(*args, **kwargs)
            except HTTPErrorCode as exc:
                total_sleep_time, request_nb = exponential_backoff(exc, max_sleep_time, multiplier,
                                                                   randomization_factor, request_nb, retry_interval,
                                                                   total_sleep_time)
                last_exception = exc
            except RequestException as exc:
                total_sleep_time, request_nb = exponential_backoff(exc, max_sleep_time, multiplier,
                                                                   randomization_factor, request_nb, retry_interval,
                                                                   total_sleep_time)
                last_exception = exc
        logger.error('Max sleep time exceeded, raising exception.')
        raise last_exception

    def exponential_backoff(exc, max_sleep_time, multiplier, randomization_factor, request_nb,
                            retry_interval, total_sleep_time):
        """Implements the exponential backoff for HTTP Requests.

        Args:
            exc:
            max_sleep_time:
            multiplier:
            randomization_factor:
            request_nb:
            retry_interval:
            total_sleep_time:

        Returns:

        """
        # Inspired from https://developers.google.com/api-client-library/java/google-http-java-client/reference/
        # 1.20.0/com/google/api/client/util/ExponentialBackOff
        next_retry_sleep = (multiplier ** request_nb *
                            (retry_interval *
                             (randint(0, int(2 * randomization_factor * 1000)) / 1000 +
                              1 - randomization_factor)))
        total_sleep_time += next_retry_sleep
        request_nb += 1
        time.sleep(next_retry_sleep)
        logger.warning('Got an exception: {exc}. Slept ({retry} seconds / {total} seconds)'
                       .format(exc=exc,
                               retry=total_sleep_time,
                               total=max_sleep_time))
        return total_sleep_time, request_nb

    # Keep the doc
    if call.__doc__:
        _retry_http.__doc__ += call.__doc__

    return _retry_http


@retry_http
def request_factory(host, req_type, endpoint, headers=None, body=None, params=None, auth=None,
                    return_none_on_404=True):
    """Makes HTTP Requests

    Args:
        host (str): Host of the URL to send.
        headers: dictionary of headers to send.
        req_type (str): HTTP verb
        endpoint: Endpoint of the URL to send.
        body: the body to attach to the request.
        params: dictionary of URL parameters to append to the URL.
        auth: Auth handler or (user, pass) tuple.
        return_none_on_404: Return None if the response has status_code 404

    Returns:
        dict: The decoded response or None if status code is 204

    Raises:
        HTTPErrorCode: Request error.
    """
    # We declare the furl host here because when 'join' is called, ``fhost`` is updated
    fhost = furl.furl(host)

    if req_type == 'GET':
        response = requests.get(fhost.join(endpoint), json=body, params=params, headers=headers, auth=auth)
    elif req_type == 'POST':
        response = requests.post(fhost.join(endpoint), json=body, params=params, headers=headers, auth=auth)
    elif req_type == 'PUT':
        response = requests.put(fhost.join(endpoint), json=body, params=params, headers=headers, auth=auth)
    elif req_type == 'DELETE':
        response = requests.delete(fhost.join(endpoint), json=body, params=params, headers=headers, auth=auth)
    else:
        raise HTTPErrorCode({'request_factory': 'Unknown request type: {0}'.format(req_type)})

    if return_none_on_404 and response.status_code == 404:
        return None
    elif response.status_code not in SUCCESS_HTTP_CODE:
        raise HTTPErrorCode({
            'error': response.status_code,
            'msg': response.text,
        })
    elif response.status_code == 204:
        return
    else:
        return response.json()



#!/usr/bin/env python

from setuptools import find_packages


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from pip.req import parse_requirements

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = [str(i.req) for i in parse_requirements("requirements.txt", session=False)]
test_requirements = [str(i.req) for i in parse_requirements("test_requirements.txt", session=False)]

VERSION = "1"

setup(
    name="pypicloud_client",
    version=str(VERSION),
    description="Provide a client for pypicloud",
    long_description=readme,
    author="Damien Goldenberg",
    author_email="damdam.gold@gmail.com",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    install_requires=requirements,
    zip_safe=False,
    keywords="pypicloud, client",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Natural Language :: English",
        "Programming Language :: Python :: 2.7",
    ],
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com"
)



import os

from pypicloud_client.http import request_factory


class PypicloudClient(object):
    """Pypicloud API client, the client can e set ny environment variables.

    Args:
        kwargs (dict) :
            user (str): User for basic authentication
            password (str): Password for basic authentication
            host (str): Pypicloud host

    """
    def __init__(self, **kwargs):
        self.host = kwargs.get("host", os.environ.get("PYPICLOUD_HOST", None))
        self.auth = (
            kwargs.get("user", os.environ.get("PYPICLOUD_USER", None)),
            kwargs.get("password", os.environ.get("PYPICLOUD_PASSWORD", None))
        )

        if self.host is None:
            raise ValueError("host have to be set, not equal to {0}".format(self.host))

        if None in self.auth:
            raise ValueError("auth have to be set, not equal to {0}".format(self.auth))

    def get_versions(self, package):
        """Get the wheel versions of an package, ordered by date from most recent to least recent.

        Args:
            package: Package whose versions to get

        Returns:
            list of dict: List of package info.
        """
        endpoint = '/api/package/{package}/'.format(package=package)
        versions = request_factory(self.host, 'GET', endpoint, auth=self.auth)
        return sorted(versions['packages'], key=lambda package: package['last_modified'], reverse=True)
